/**
 * FLL - Level 0
 *
 * Project: Time
 * API Version: 0.7
 * Licenses: lgpl-2.1-or-later
 *
 * Provides time related functionality.
 */
#ifndef _F_time_h
#define _F_time_h

// Libc includes.
#include <time.h>
#include <sys/time.h>

// FLL-0 includes.
#include <fll/level_0/type.h>
#include <fll/level_0/type/time.h>
#include <fll/level_0/status.h>
#include <fll/level_0/string.h>

// FLL-0 time includes.
#include <fll/level_0/time/common.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Convert the given time into a calendar-time string.
 *
 * @param time
 *   The time to convert.
 * @param destination
 *   The calendar string representation of the given time.
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_buffer_overflow (with error bit) on buffer overflow.
 *
 *   Errors (with error bit) from: f_string_append().
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see ctime_r()
 *
 * @see f_string_append()
 */
#ifndef _di_f_time_calendar_string_
  extern f_status_t f_time_calendar_string(const time_t time, f_string_dynamic_t * const destination);
#endif // _di_f_time_calendar_string_

/**
 * Convert the given time into a calendar-time string from a broken up (part) date and time.
 *
 * @param part
 *   The time parts to convert.
 * @param destination
 *   The calendar string representation of the given time.
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_buffer_overflow (with error bit) on buffer overflow.
 *
 *   Errors (with error bit) from: f_string_append().
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see asctime_r()
 *
 * @see f_string_append()
 */
#ifndef _di_f_time_calendar_string_part_
  extern f_status_t f_time_calendar_string_part(const f_time_part_t part, f_string_dynamic_t * const destination);
#endif // _di_f_time_calendar_string_part_

/**
 * Get the current seconds and nanoseconds from the system clock.
 *
 * @param code
 *   Designates how to get the time from the system clock.
 * @param time
 *   The time retrieved.
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_access_denied (with error bit) on access denied.
 *   F_buffer (with error bit) if the buffer is invalid.
 *   F_buffer_overflow (with error bit) if the time stamp does not fit in the time range.
 *   F_device_not (with error bit) if the system clock is no onger available (such as when a hot-pluggable clock disappears).
 *   F_parameter (with error bit) if a parameter is invalid.
 *   F_prohibited (with error bit) if not permitted to perform this operation.
 *   F_support_not (with error bit) if the operation specified by the code is not supported by the clock.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see clock_gettime()
 */
#ifndef _di_f_time_clock_get_
  extern f_status_t f_time_clock_get(const clockid_t code, f_time_spec_t * const time);
#endif // _di_f_time_clock_get_

/**
 * Get the precision from the system clock.
 *
 * @param code
 *   Designates how to get the time from the system clock.
 * @param precision
 *   The clock precision (clock resolution).
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_access_denied (with error bit) on access denied.
 *   F_buffer (with error bit) if the buffer is invalid.
 *   F_buffer_overflow (with error bit) if the time stamp does not fit in the time range.
 *   F_device_not (with error bit) if the system clock is no onger available (such as when a hot-pluggable clock disappears).
 *   F_parameter (with error bit) if a parameter is invalid.
 *   F_prohibited (with error bit) if not permitted to perform this operation.
 *   F_support_not (with error bit) if the operation specified by the code is not supported by the clock.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see clock_getres()
 */
#ifndef _di_f_time_clock_precision_
  extern f_status_t f_time_clock_precision(const clockid_t code, f_time_spec_t * const precision);
#endif // _di_f_time_clock_precision_

/**
 * Set the current seconds and nanoseconds from the system clock.
 *
 * @param code
 *   Designates how to set the time of the system clock.
 * @param time
 *   The time to set.
 *
 * @return
 *   F_okay on success.
 *
 *   F_access_denied (with error bit) on access denied.
 *   F_buffer (with error bit) if the buffer is invalid.
 *   F_buffer_overflow (with error bit) if the time stamp does not fit in the time range.
 *   F_device_not (with error bit) if the system clock is no onger available (such as when a hot-pluggable clock disappears).
 *   F_parameter (with error bit) if a parameter is invalid.
 *   F_prohibited (with error bit) if not permitted to perform this operation.
 *   F_support_not (with error bit) if the operation specified by the code is not supported by the clock.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see clock_settime()
 */
#ifndef _di_f_time_clock_set_
  extern f_status_t f_time_clock_set(const clockid_t code, const f_time_spec_t time);
#endif // _di_f_time_clock_set_

/**
 * Get the current time in seconds since epoch.
 *
 * @param time
 *   The time retrieved.
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_buffer (with error bit) if the buffer is invalid.
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see time()
 */
#ifndef _di_f_time_epoch_get_
  extern f_status_t f_time_epoch_get(time_t * const time);
#endif // _di_f_time_epoch_get_

/**
 * Convert the given time into local time parts.
 *
 * @param time
 *   The time to convert.
 * @param destination
 *   The time parts.
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_buffer_overflow (with error bit) on buffer overflow.
 *
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see localtime_r()
 */
#ifndef _di_f_time_local_get_
  extern f_status_t f_time_local_get(const time_t time, f_time_part_t * const destination);
#endif // _di_f_time_local_get_

/**
 * Get the current time of day.
 *
 * @param time
 *   The time retrieved.
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_buffer (with error bit) if the buffer is invalid.
 *   F_parameter (with error bit) if a parameter is invalid.
 *   F_prohibited (with error bit) if the file system does not permit this operation.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see gettimeofday()
 */
#ifndef _di_f_time_of_day_get_
  extern f_status_t f_time_of_day_get(f_time_value_t * const time);
#endif // _di_f_time_of_day_get_

/**
 * Set the current time of day.
 *
 * @param time
 *   The time to set.
 *
 * @return
 *   F_okay on success.
 *
 *   F_number_overflow (with error bit) if the summation of seconds and nanoseconds results in an overflow when cast to appropriate integer type.
 *   F_number_underflow (with error bit) if the second or nanoseconds is less than 0 when cast to appropriate integer type.
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see settimeofday()
 */
#ifndef _di_f_time_of_day_set_
  extern f_status_t f_time_of_day_set(const f_time_value_t time);
#endif // _di_f_time_of_day_set_

/**
 * Sleep a given amount of time specified by the f_time_spec_t.
 *
 * @param time
 *   The time to sleep (in seconds and nanoseconds).
 * @param remaining
 *   (optional) Contains the remaining time if stopped on interrupt.
 *
 *   Set to NULL to not use.
 *
 * @return
 *   F_okay on success.
 *   F_interrupt if interrupted by a signal.
 *
 *   F_buffer (with error bit) if failed to copy data from user space.
 *   F_parameter (with error bit) if a parameter is invalid (such as tv_nsec is not in range 0 to 999999999 or tv_sec is negative).
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see nanosleep()
 */
#ifndef _di_f_time_sleep_spec_
  extern f_status_t f_time_sleep_spec(const f_time_spec_t time, f_time_spec_t * const remaining);
#endif // _di_f_time_sleep_spec_

/**
 * Create a timespec representing the given seconds and milliseconds.
 *
 * The "f_time_spec_t" may have different lengths and so this function provides overflow and underflow protection.
 * The "f_time_simple_t" should not need this as "f_time_simple_t" is intended to always be uint64_t.
 *
 * @param second
 *   The number of seconds.
 * @param millisecond
 *   The number of milliseconds.
 * @param time
 *   The time spec representing the given seconds and milliseconds.
 *
 *   This is still updated on F_number_overflow, and the result of the overflow is still assigned to seconds.
 *   This is still updated on F_number_underflow, but the underflow value is set to 0.
 *
 * @return
 *   F_okay on success.
 *
 *   F_number_overflow (with error bit) if the summation of seconds and milliseconds results in an overflow when cast to appropriate integer type.
 *   F_number_underflow (with error bit) if the seconds or milliseconds is less than 0 when cast to appropriate integer type.
 *   F_parameter (with error bit) if a parameter is invalid.
 */
#ifndef _di_f_time_spec_millisecond_
  extern f_status_t f_time_spec_millisecond(const f_number_unsigned_t second, const f_number_unsigned_t millisecond, f_time_spec_t * const time);
#endif // _di_f_time_spec_millisecond_

/**
 * Create a timespec representing the given seconds and nanoseconds.
 *
 * The "f_time_spec_t" may have different lengths and so this function provides overflow and underflow protection.
 * The "f_time_simple_t" should not need this as "f_time_simple_t" is intended to always be uint64_t.
 *
 * @param second
 *   The number of seconds.
 * @param nanosecond
 *   The number of nanoseconds.
 * @param time
 *   The time spec representing the given seconds and nanoseconds.
 *
 *   This is still updated on F_number_overflow, and the result of the overflow is still assigned to second.
 *   This is still updated on F_number_underflow, but the underflow value is set to 0.
 *
 * @return
 *   F_okay on success.
 *
 *   F_number_overflow (with error bit) if the summation of seconds and nanoseconds results in an overflow when cast to appropriate integer type.
 *   F_number_underflow (with error bit) if the second or nanoseconds is less than 0 when cast to appropriate integer type.
 *   F_parameter (with error bit) if a parameter is invalid.
 */
#ifndef _di_f_time_spec_nanosecond_
  extern f_status_t f_time_spec_nanosecond(const f_number_unsigned_t second, const f_number_unsigned_t nanosecond, f_time_spec_t * const time);
#endif // _di_f_time_spec_nanosecond_

/**
 * Convert the given time into UTC time parts.
 *
 * @param time
 *   The time to convert.
 * @param destination
 *   The time parts.
 *
 *   Must not be NULL.
 *
 * @return
 *   F_okay on success.
 *
 *   F_buffer_overflow (with error bit) on buffer overflow.
 *
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 *   F_failure (with error bit) on any other error.
 *
 * @see gmtime_r()
 */
#ifndef _di_f_time_utc_get_
  extern f_status_t f_time_utc_get(const time_t time, f_time_part_t * const destination);
#endif // _di_f_time_utc_get_

#ifdef __cplusplus
} // extern "C"
#endif

#endif // _F_time_h
