/**
 * FLL - Level 1
 *
 * Project: FSS
 * API Version: 0.7
 * Licenses: lgpl-2.1-or-later
 *
 * These are provided for internal reduction in redundant code.
 * These should not be exposed/used outside of this project.
 */
#ifndef _PRIVATE_FL_fss_list_h
#define _PRIVATE_FL_fss_list_h

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Read a standard list Object where the list open is specific to a given FSS standard.
 *
 * This will record where delimit placeholders exist but will not apply the delimits.
 *
 * @param list_open
 *   The list open character for the given standard.
 *   This must be defined where list_open.used = 1.
 * @param buffer
 *   The buffer to read from.
 * @param range
 *   The start/stop location within the buffer to be processed.
 *   The start location will be updated as the buffer is being processed.
 *   The start location will represent where the read stopped on return.
 *   A start location past the stop location or buffer used means that the entire range was processed.
 *
 *   Must not be NULL.
 * @param found
 *   A location where a valid Object is found.
 *
 *   Must not be NULL.
 * @param delimits
 *   A delimits array representing where delimits exist within the buffer.
 *
 *   Must not be NULL.
 * @param state
 *   A state for providing flags and handling interrupts during long running operations.
 *   There is no state.handle().
 *   There is no "callbacks" structure.
 *   There is no data structure passed to these functions.
 *
 *   When state.interrupt() returns, only F_interrupt and F_interrupt_not are processed.
 *   Error bit designates an error but must be passed along with F_interrupt.
 *   All other statuses are ignored.
 *
 *   Must not be NULL.
 *
 *   This alters state.status:
 *     F_fss_found_object on success and Object is found (start location is at end of Object).
 *     F_fss_found_object_not on success and no Object is found (start location is after character designating this is not an Object).
 *     F_okay_eos on success after reaching the end of the buffer (a valid Object is not yet confirmed).
 *     F_okay_stop on success after reaching stopping point (a valid Object is not yet confirmed).
 *     F_data_not if buffer is empty (buffer.used is 0).
 *     F_data_not_eos no objects found after reaching the end of the buffer (essentially only comments are found).
 *     F_data_not_stop no data found after reaching stopping point (essentially only comments are found).
 *     F_end_not_group_eos if EOS was reached before the a group termination was reached.
 *     F_end_not_group_stop if stop point was reached before the a group termination was reached.
 *
 *     F_interrupt (with error bit) if stopping due to an interrupt.
 *     F_parameter (with error bit) if a parameter is invalid.
 *
 *     Errors (with error bit) from: f_fss_is_graph().
 *     Errors (with error bit) from: f_fss_is_space().
 *     Errors (with error bit) from: f_fss_is_zero_width().
 *     Errors (with error bit) from: f_fss_seek_to_eol().
 *     Errors (with error bit) from: f_fss_skip_past_delimit().
 *     Errors (with error bit) from: f_fss_skip_past_space().
 *     Errors (with error bit) from: f_memory_array_increase_by().
 *     Errors (with error bit) from: f_utf_buffer_increment().
 *
 * @see f_fss_is_graph()
 * @see f_fss_is_space()
 * @see f_fss_is_zero_width()
 * @see f_fss_seek_to_eol()
 * @see f_fss_skip_past_delimit()
 * @see f_fss_skip_past_space()
 * @see f_memory_array_increase_by()
 * @see f_utf_buffer_increment()
 *
 * @see fl_fss_basic_list_object_read()
 * @see fl_fss_embedded_list_object_read()
 * @see fl_fss_extended_list_object_read()
 */
#if !defined(_di_fl_fss_basic_list_object_read_) || !defined(_di_fl_fss_embedded_list_object_read_) || !defined(_di_fl_fss_extended_list_object_read_)
  extern void private_fl_fss_list_object_read(const f_string_static_t list_open, const f_string_static_t buffer, f_range_t * const range, f_range_t * const found, f_number_unsigneds_t * const delimits, f_state_t * const state) F_attribute_visibility_internal_d;
#endif // !defined(_di_fl_fss_basic_list_object_read_) || !defined(_di_fl_fss_embedded_list_object_read_) || !defined(_di_fl_fss_extended_list_object_read_)

/**
 * Add all bytes to destination until stop point, buffer end, or EOL.
 *
 * @param buffer
 *   The buffer to seek through.
 * @param range
 *   The start/stop location within the buffer string to process.
 * @param destination
 *   The buffer where the bytes are written to.
 * @param state
 *   A state for providing flags and handling interrupts during long running operations.
 *   There is no state.handle().
 *   There is no "callbacks" structure.
 *   There is no data structure passed to these functions.
 *
 *   When state.interrupt() returns, only F_interrupt and F_interrupt_not are processed.
 *   Error bit designates an error but must be passed along with F_interrupt.
 *   All other statuses are ignored.
 *
 *   This alters state.status:
 *     F_okay on success.
 *
 *     F_interrupt (with error bit) if stopping due to an interrupt.
 *
 *     Errors (with error bit) from: f_memory_array_increase().
 *
 * @see f_memory_array_increase()
 * @see fl_fss_basic_list_content_write()
 * @see fl_fss_embedded_list_content_write()
 * @see fl_fss_extended_list_content_write()
 */
#if !defined(_di_fl_fss_basic_list_content_write_) || !defined(_di_fl_fss_extended_list_content_write_) || !defined(_di_fl_fss_embedded_list_content_write_)
  extern void private_fl_fss_list_write_add_until_end(const f_string_static_t buffer, f_range_t * const range, f_string_dynamic_t * const destination, f_state_t * const state) F_attribute_visibility_internal_d;
#endif // !defined(_di_fl_fss_basic_list_content_write_) || !defined(_di_fl_fss_extended_list_content_write_) || !defined(_di_fl_fss_embedded_list_content_write_)

/**
 * Trim a given object used by the basic list and extended list object write functions.
 *
 * @param used_start
 *   The destination.used value before any operations were performed.
 * @param destination
 *   The buffer where the object is written to.
 * @param state
 *   A state for providing flags and handling interrupts during long running operations.
 *   There is no state.handle().
 *   There is no "callbacks" structure.
 *   There is no data structure passed to these functions.
 *
 *   When state.interrupt() returns, only F_interrupt and F_interrupt_not are processed.
 *   Error bit designates an error but must be passed along with F_interrupt.
 *   All other statuses are ignored.
 *
 *   This alters state.status:
 *     F_okay on success.
 *
 *     F_interrupt (with error bit) if stopping due to an interrupt.
 *
 *     Errors (with error bit) from: f_fss_is_space().
 *
 * @see f_fss_is_space()
 * @see fl_fss_basic_list_object_write()
 * @see fl_fss_extended_list_object_write()
 */
#if !defined(_di_fl_fss_basic_list_object_write_) || !defined(_di_fl_fss_extended_list_object_write_)
  extern void private_fl_fss_list_write_object_trim(const f_number_unsigned_t used_start, f_string_dynamic_t * const destination, f_state_t * const state) F_attribute_visibility_internal_d;
#endif // !defined(_di_fl_fss_basic_list_object_write_) || !defined(_di_fl_fss_extended_list_object_write_)

#ifdef __cplusplus
} // extern "C"
#endif

#endif // _PRIVATE_FL_fss_list_h
